from collections import namedtuple
from math import isclose
from pathlib import Path
from typing import Set

import SimpleITK
import numpy as np
import pydicom

from panimg.image_builders.utils import convert_itk_to_internal
from panimg.types import ImageBuilderResult

NUMPY_IMAGE_TYPES = {
    "character": SimpleITK.sitkUInt8,
    "uint8": SimpleITK.sitkUInt8,
    "uint16": SimpleITK.sitkUInt16,
    "uint32": SimpleITK.sitkUInt32,
    "uint64": SimpleITK.sitkUInt64,
    "int8": SimpleITK.sitkInt8,
    "int16": SimpleITK.sitkInt16,
    "int32": SimpleITK.sitkInt32,
    "int64": SimpleITK.sitkInt64,
    "float32": SimpleITK.sitkFloat32,
    "float64": SimpleITK.sitkFloat64,
}

OPTIONAL_METADATA_FIELDS = (
    # These fields will be included in the output mha file
    "Laterality",
    "PatientID",
    "PatientName",
    "PatientBirthDate",
    "PatientAge",
    "PatientSex",
    "StudyDate",
    "StudyInstanceUID",
    "SeriesInstanceUID",
    "StudyDescription",
    "SliceThickness",
    "WindowCenter",
    "WindowWidth",
)


def pixel_data_reached(tag, vr, length):
    return pydicom.datadict.keyword_for_tag(tag) == "PixelData"


def _get_headers_by_study(files: Set[Path]):
    """
    Gets all headers from dicom files found in path.

    Parameters
    ----------
    path
        Path to a directory that contains all images that were uploaded during
        an upload session.

    Returns
    -------
    A dictionary of sorted headers for all dicom image files found within path,
    grouped by study id.
    """
    studies = {}
    errors = {}
    indices = {}

    for file in files:
        if not file.is_file():
            continue
        with file.open("rb") as f:
            try:
                ds = pydicom.filereader.read_partial(
                    f, stop_when=pixel_data_reached
                )
                dims = f"{ds.Rows}x{ds.Columns}"
                key = f"{ds.StudyInstanceUID}-{dims}"
                studies[key] = studies.get(key, {})
                indices[ds.StudyInstanceUID] = indices.get(
                    ds.StudyInstanceUID, {}
                )
                index = indices[ds.StudyInstanceUID].get(dims)
                if index is None:
                    index = (
                        max(list(indices[ds.StudyInstanceUID].values()) + [-1])
                        + 1
                    )
                    indices[ds.StudyInstanceUID][dims] = index
                headers = studies[key].get("headers", [])
                headers.append({"file": file, "data": ds})
                studies[key]["index"] = index
                studies[key]["headers"] = headers
            except Exception as e:
                errors[file] = format_error(e)

    for key in studies:
        studies[key]["headers"].sort(
            key=lambda x: int(x["data"].InstanceNumber)
        )
    return studies, errors


def format_error(message):
    return f"Dicom image builder: {message}"


def _validate_dicom_files(files: Set[Path]):
    """
    Gets the headers for all dicom files on path and validates them.

    Parameters
    ----------
    path
        Path to a directory that contains all images that were uploaded during
        an upload session.

    Returns
    -------
    A list of `dicom_dataset` named tuples per study, consisting of:
     - Headers for all dicom image files for the study
     - Number of time points
     - Number of slices per time point

    Any study with an inconsistent amount of slices per time point is discarded.
    """
    studies, errors = _get_headers_by_study(files)
    result = []
    dicom_dataset = namedtuple(
        "dicom_dataset", ["headers", "n_time", "n_slices", "index"]
    )
    for key in studies:
        headers = studies[key]["headers"]
        index = studies[key]["index"]
        if not headers:
            continue
        data = headers[-1]["data"]
        n_time = getattr(data, "TemporalPositionIndex", None)
        n_slices = max(len(headers), int(getattr(data, "NumberOfFrames", 0)))
        # Not a 4d dicom file
        if n_time is None:
            result.append(
                dicom_dataset(
                    headers=headers,
                    n_time=n_time,
                    n_slices=n_slices,
                    index=index,
                )
            )
            continue
        if len(headers) % n_time > 0:
            for d in headers:
                errors[d["file"]] = format_error(
                    "Number of slices per time point differs"
                )
            continue
        n_slices = n_slices // n_time
        result.append(
            dicom_dataset(
                headers=headers, n_time=n_time, n_slices=n_slices, index=index,
            )
        )
    del studies
    return result, errors


def _extract_direction(dicom_ds, direction):
    try:
        # Try to extract the direction from the file
        sitk_ref = SimpleITK.ReadImage(str(dicom_ds.headers[0]["file"]))
        # The direction per slice is a 3x3 matrix, so we add the time
        # dimension ourselves
        dims = sitk_ref.GetDimension()
        _direction = np.reshape(sitk_ref.GetDirection(), (dims, dims))
        direction[:dims, :dims] = _direction
    except Exception:
        pass
    return direction


def _process_dicom_file(  # noqa: C901
    *, dicom_ds, created_image_prefix, output_directory
):
    ref_file = pydicom.dcmread(str(dicom_ds.headers[0]["file"]))
    ref_origin = tuple(
        float(i) for i in getattr(ref_file, "ImagePositionPatient", (0, 0, 0))
    )
    dimensions = 4 if dicom_ds.n_time and dicom_ds.n_time > 1 else 3
    direction = np.eye(dimensions, dtype=float)
    direction = _extract_direction(dicom_ds, direction)
    pixel_dims = (
        dicom_ds.n_slices,
        int(ref_file.Rows),
        int(ref_file.Columns),
    )
    if dimensions == 4:
        pixel_dims = (dicom_ds.n_time,) + pixel_dims

    # Additional Meta data Contenttimes and Exposures
    content_times = []
    exposures = []
    origin = None
    origin_diff = np.array((0, 0, 0), dtype=float)
    n_diffs = 0
    for partial in dicom_ds.headers:
        ds = partial["data"]
        if "ImagePositionPatient" in ds:
            file_origin = np.array(ds.ImagePositionPatient, dtype=float)
            if origin is not None:
                diff = file_origin - origin
                origin_diff = origin_diff + diff
                n_diffs += 1
            origin = file_origin
    avg_origin_diff = tuple(origin_diff / n_diffs)
    try:
        z_i = avg_origin_diff[2]
    except IndexError:
        z_i = 1.0

    samples_per_pixel = int(getattr(ref_file, "SamplesPerPixel", 1))
    img = _create_itk_from_dcm(
        content_times=content_times,
        dicom_ds=dicom_ds,
        dimensions=dimensions,
        exposures=exposures,
        pixel_dims=pixel_dims,
        z_i=z_i,
        samples_per_pixel=samples_per_pixel,
    )

    if origin is None:
        origin = (0.0, 0.0, 0.0)
    sitk_origin = ref_origin if z_i >= 0.0 else tuple(origin)
    z_i = np.abs(z_i) if not np.isnan(z_i) else 1.0

    if "PixelSpacing" in ref_file:
        x_i, y_i = (float(x) for x in ref_file.PixelSpacing)
    else:
        x_i = y_i = 1.0

    sitk_spacing = (x_i, y_i, z_i)
    if dimensions == 4:
        sitk_spacing += (1.0,)
        sitk_origin += (0.0,)

    sitk_direction = tuple(direction.flatten())
    img.SetDirection(sitk_direction)
    img.SetSpacing(sitk_spacing)
    img.SetOrigin(sitk_origin)

    if dimensions == 4:
        # Set Additional Meta Data
        img.SetMetaData("ContentTimes", " ".join(content_times))
        img.SetMetaData("Exposures", " ".join(exposures))

    for f in OPTIONAL_METADATA_FIELDS:
        if getattr(ref_file, f, False):
            img.SetMetaData(f, str(getattr(ref_file, f)))

    # Convert the SimpleITK image to our internal representation
    return convert_itk_to_internal(
        simple_itk_image=img,
        name=f"{created_image_prefix}-{dicom_ds.headers[0]['data'].StudyInstanceUID}-{dicom_ds.index}",
        output_directory=output_directory,
    )


def _create_itk_from_dcm(
    *,
    content_times,
    dicom_ds,
    dimensions,
    exposures,
    pixel_dims,
    z_i,
    samples_per_pixel,
):
    apply_slope = any(
        not isclose(float(getattr(h["data"], "RescaleSlope", 1.0)), 1.0)
        for h in dicom_ds.headers
    )
    apply_intercept = any(
        not isclose(float(getattr(h["data"], "RescaleIntercept", 0.0)), 0.0)
        for h in dicom_ds.headers
    )
    apply_scaling = apply_slope or apply_intercept
    is_rgb = samples_per_pixel > 1
    if apply_scaling:
        np_dtype = np.float32
    else:
        np_dtype = np.short
    if samples_per_pixel > 1:
        pixel_dims += (samples_per_pixel,)
    dcm_array = None
    use_pixel_array = False

    for index, partial in enumerate(dicom_ds.headers):
        ds = pydicom.dcmread(str(partial["file"]))

        if apply_scaling:
            pixel_array = float(
                getattr(ds, "RescaleSlope", 1)
            ) * ds.pixel_array + float(getattr(ds, "RescaleIntercept", 0))
        else:
            pixel_array = ds.pixel_array

        if (
            len(pixel_array.shape) == dimensions
            or pixel_array.shape == pixel_dims
        ):
            use_pixel_array = True
            del ds
            break
        if dcm_array is None:
            dcm_array = np.zeros(pixel_dims, dtype=np_dtype)

        z_index = index if z_i >= 0 else len(dicom_ds.headers) - index - 1

        if dimensions == 4:
            dcm_array[
                index // dicom_ds.n_slices, z_index % dicom_ds.n_slices, :, :
            ] = pixel_array
            if index % dicom_ds.n_slices == 0:
                content_times.append(str(ds.ContentTime))
                exposures.append(str(ds.Exposure))
        else:
            dcm_array[z_index % dicom_ds.n_slices, :, :] = pixel_array

        del ds
    if use_pixel_array:
        img = SimpleITK.GetImageFromArray(pixel_array, isVector=is_rgb)
    else:
        img = SimpleITK.GetImageFromArray(dcm_array, isVector=is_rgb)
    return img


def image_builder_dicom(
    *, files: Set[Path], output_directory: Path, created_image_prefix: str = ""
) -> ImageBuilderResult:
    """
    Constructs image objects by inspecting files in a directory.

    Parameters
    ----------
    path
        Path to a directory that contains all images that were uploaded during
        an upload session.

    Returns
    -------
    An `ImageBuilder` object consisting of:
     - a list of filenames for all files consumed by the image builder
     - a list of detected images
     - a list files associated with the detected images
     - path->error message map describing what is wrong with a given file
    """
    studies, file_errors = _validate_dicom_files(files)
    new_images = set()
    new_image_files = set()
    consumed_files = set()
    for dicom_ds in studies:
        try:
            n_image, n_image_files = _process_dicom_file(
                dicom_ds=dicom_ds,
                created_image_prefix=created_image_prefix,
                output_directory=output_directory,
            )
            new_images.add(n_image)
            new_image_files |= set(n_image_files)
            consumed_files |= {d["file"] for d in dicom_ds.headers}
        except Exception as e:
            for d in dicom_ds.headers:
                file_errors[d["file"]] = format_error(e)

    return ImageBuilderResult(
        consumed_files=consumed_files,
        file_errors=file_errors,
        new_images=new_images,
        new_image_files=new_image_files,
        new_folders=set(),
    )



import os
import re
import shutil
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List, Optional, Set
from uuid import UUID, uuid4

import openslide
import pyvips
import tifffile

from panimg.exceptions import ValidationError
from panimg.models import (
    ColorSpace,
    ImageType,
    PanImg,
    PanImgFile,
    PanImgFolder,
)
from panimg.settings import DZI_TILE_SIZE
from panimg.types import ImageBuilderResult


@dataclass
class GrandChallengeTiffFile:
    path: Path
    pk: UUID = field(default_factory=uuid4)
    image_width: int = 0
    image_height: int = 0
    resolution_levels: int = 0
    color_space: Optional[ColorSpace] = None
    voxel_width_mm: float = 0
    voxel_height_mm: float = 0
    voxel_depth_mm: Optional[float] = None
    source_files: List = field(default_factory=list)
    associated_files: List = field(default_factory=list)

    def validate(self):
        if not self.image_width:
            raise ValidationError(
                "Not a valid tif: Image width could not be determined"
            )
        if not self.image_height:
            raise ValidationError(
                "Not a valid tif: Image height could not be determined"
            )
        if not self.resolution_levels:
            raise ValidationError(
                "Not a valid tif: Resolution levels not valid"
            )
        if not self.voxel_width_mm:
            raise ValidationError(
                "Not a valid tif: Voxel width could not be determined"
            )
        if not self.voxel_height_mm:
            raise ValidationError(
                "Not a valid tif: Voxel height could not be determined"
            )


def _get_tag_value(tags, tag):
    try:
        return tags[tag].value
    except (KeyError, AttributeError):
        return None


def _get_voxel_spacing_mm(tags, tag):
    """
    Calculate the voxel spacing in mm.

    Use the set of tags from the tiff image to calculate the spacing for a
    particular dimension. Supports INCH and CENTIMETER resolution units.

    Parameters
    ----------
    tags
        The collection of tags from the tif file.
    tag
        The tag that contains the resolution of the tif file along the
        dimension of interest.

    Raises
    ------
    ValidationError
        Raised if an unrecognised resolution unit is used.


    Returns
    -------
        The voxel spacing in mm.

    """
    try:
        # Resolution is a tuple of the number of pixels and the length of the
        # image in cm or inches, depending on resolution unit
        resolution_unit = str(_get_tag_value(tags, "ResolutionUnit"))
        resolution = _get_tag_value(tags, tag)
        if resolution_unit == "RESUNIT.INCH":
            return 25.4 / (resolution[0] / resolution[1])
        elif resolution_unit == "RESUNIT.CENTIMETER":
            return 10 / (resolution[0] / resolution[1])
        raise ValidationError(
            f"Invalid resolution unit {resolution_unit}" f" in tiff file"
        )
    except (ZeroDivisionError, TypeError, IndexError):
        raise ValidationError("Invalid resolution in tiff file")


def _extract_openslide_properties(
    *, gc_file: GrandChallengeTiffFile, image: any
) -> GrandChallengeTiffFile:
    if not gc_file.voxel_width_mm and "openslide.mpp-x" in image.properties:
        gc_file.voxel_width_mm = (
            float(image.properties["openslide.mpp-x"]) / 1000
        )
    if not gc_file.voxel_height_mm and "openslide.mpp-y" in image.properties:
        gc_file.voxel_height_mm = (
            float(image.properties["openslide.mpp-y"]) / 1000
        )
    if (
        not gc_file.image_height
        and "openslide.level[0].height" in image.properties
    ):
        gc_file.image_height = int(
            image.properties["openslide.level[0].height"]
        )

    if (
        not gc_file.image_width
        and "openslide.level[0].width" in image.properties
    ):
        gc_file.image_width = int(image.properties["openslide.level[0].width"])

    if (
        not gc_file.resolution_levels
        and "openslide.level-count" in image.properties
    ):
        gc_file.resolution_levels = int(
            image.properties["openslide.level-count"]
        )
    return gc_file


def _extract_tags(
    *, gc_file: GrandChallengeTiffFile, pages: tifffile.tifffile.TiffPages
) -> GrandChallengeTiffFile:
    """
    Extracts tags form a tiff file loaded with tifffile for use in grand challenge
    :param pages: The pages and tags from tifffile
    :return: The GrandChallengeTiffFile with the properties defined in the tags
    """
    if not pages:
        return gc_file

    tags = pages[0].tags

    gc_file.resolution_levels = len(pages)

    gc_file.color_space = _get_color_space(
        color_space_string=str(
            _get_tag_value(tags, "PhotometricInterpretation")
        )
    )
    gc_file.image_width = _get_tag_value(tags, "ImageWidth")
    gc_file.image_height = _get_tag_value(tags, "ImageLength")

    #  some formats like the Philips tiff don't have the spacing in their tags,
    #  we retrieve them later with OpenSlide
    if "XResolution" in tags:
        gc_file.voxel_width_mm = _get_voxel_spacing_mm(tags, "XResolution")
        gc_file.voxel_height_mm = _get_voxel_spacing_mm(tags, "YResolution")

    gc_file.voxel_depth_mm = None

    return gc_file


def _get_color_space(*, color_space_string) -> Optional[ColorSpace]:
    color_space_string = color_space_string.split(".")[1].upper()

    if color_space_string == "MINISBLACK":
        color_space = ColorSpace.GRAY
    else:
        try:
            color_space = ColorSpace[color_space_string]
        except KeyError:
            return None

    return color_space


def _create_image_file(
    *, path: Path, image: PanImg, output_directory: Path
) -> PanImgFile:

    output_file = output_directory / f"{image.pk}{path.suffix}"

    if path.suffix.lower() == ".dzi":
        image_type = ImageType.DZI
    else:
        # TODO (jmsmkn): Create the tiff files in the correct location
        shutil.copy(src=str(path.resolve()), dst=str(output_file.resolve()))
        image_type = ImageType.TIFF

    return PanImgFile(
        image_id=image.pk, image_type=image_type, file=output_file
    )


def _load_with_tiff(
    *, gc_file: GrandChallengeTiffFile
) -> GrandChallengeTiffFile:
    tiff_file = tifffile.TiffFile(str(gc_file.path.absolute()))
    gc_file = _extract_tags(gc_file=gc_file, pages=tiff_file.pages)
    return gc_file


def _load_with_openslide(
    *, gc_file: GrandChallengeTiffFile
) -> (str, GrandChallengeTiffFile):
    open_slide_file = openslide.open_slide(str(gc_file.path.absolute()))
    gc_file = _extract_openslide_properties(
        gc_file=gc_file, image=open_slide_file
    )
    return gc_file


def _new_image_files(
    *, gc_file: GrandChallengeTiffFile, image: PanImg, output_directory: Path
) -> Set[PanImgFile]:
    new_image_files = {
        _create_image_file(
            path=gc_file.path.absolute(),
            image=image,
            output_directory=output_directory,
        )
    }

    if gc_file.source_files:
        for s in gc_file.source_files:
            new_image_files.add(
                _create_image_file(
                    path=s, image=image, output_directory=output_directory
                )
            )

    return new_image_files


def _new_folder_uploads(
    *, dzi_output: Path, image: PanImg,
) -> Set[PanImgFolder]:
    new_folder_upload = set()

    if dzi_output:
        dzi_folder_upload = PanImgFolder(
            folder=dzi_output.parent / (dzi_output.name + "_files"),
            image_id=image.pk,
        )
        new_folder_upload.add(dzi_folder_upload)

    return new_folder_upload


mirax_pattern = r"INDEXFILE\s?=\s?.|FILE_\d+\s?=\s?."
# vms (and vmu) files contain key value pairs, where the ImageFile keys
# can have the following format:
# ImageFile =, ImageFile(x,y) ImageFile(z,x,y)
vms_pattern = r"ImageFile(\(\d*,\d*(,\d)?\))?\s?=\s?.|MapFile\s?=\s?.|OptimisationFile\s?=\s?.|MacroImage\s?=\s?."


def _get_mrxs_files(mrxs_file: Path):
    # Find Slidedat.ini, which provides us with all the other file names
    name, _ = os.path.splitext(mrxs_file.name)
    slide_dat = mrxs_file.parent / name / "Slidedat.ini"
    file_matched = []
    if not slide_dat.exists():
        slide_dat = [
            f
            for f in slide_dat.parent.iterdir()
            if f.name.lower() == slide_dat.name.lower()
        ][0]
    with open(slide_dat, "r") as f:
        lines = [
            line for line in f.readlines() if re.match(mirax_pattern, line)
        ]
        for line in lines:
            original_name = line.split("=")[1].strip()
            if os.path.exists(slide_dat.parent / original_name):
                file_matched.append(slide_dat.parent / original_name)
        file_matched.append(slide_dat)
    return file_matched


def _get_vms_files(vms_file: Path):
    file_matched = []
    with open(str(vms_file.absolute()), "r") as f:
        lines = [line for line in f.readlines() if re.match(vms_pattern, line)]
        for line in lines:
            original_name = line.split("=")[1].strip()
            if os.path.exists(vms_file.parent / original_name):
                file_matched.append(vms_file.parent / original_name)
        return file_matched


def _convert(
    files: List[Path], associated_files_getter: Optional[callable], converter
) -> (List[GrandChallengeTiffFile], Dict):
    compiled_files: List[GrandChallengeTiffFile] = []
    associated_files: List[Path] = []
    errors = {}
    for file in files:
        try:
            gc_file = GrandChallengeTiffFile(file)
            if associated_files_getter:
                associated_files = associated_files_getter(gc_file.path)
            tiff_file = _convert_to_tiff(
                path=file, pk=gc_file.pk, converter=converter
            )
        except Exception as e:
            errors[file] = str(e)
            continue
        else:
            gc_file.path = tiff_file
            gc_file.associated_files = associated_files
            gc_file.associated_files.append(file)
            compiled_files.append(gc_file)
    return compiled_files, errors


def _convert_to_tiff(*, path: Path, pk: UUID, converter) -> Path:
    new_file_name = path.parent / f"{path.stem}_{str(pk)}.tif"
    image = converter.Image.new_from_file(
        str(path.absolute()), access="sequential"
    )

    # correct xres and yres if they have default value of 1
    # can be removed once updated to VIPS 8.10
    if image.get("xres") == 1 and "openslide.mpp-x" in image.get_fields():
        x_res = 1000.0 / float(image.get("openslide.mpp-x"))
        y_res = 1000.0 / float(image.get("openslide.mpp-y"))
        image = image.copy(xres=x_res, yres=y_res)

    image.write_to_file(
        str(new_file_name.absolute()),
        tile=True,
        pyramid=True,
        bigtiff=True,
        compression="jpeg",
        Q=70,
    )
    return new_file_name


def _load_gc_files(
    *, files: Set[Path], converter
) -> (List[GrandChallengeTiffFile], Dict):
    loaded_files = []
    errors = {}
    complex_file_handlers = {
        ".mrxs": _get_mrxs_files,
        ".vms": _get_vms_files,
        ".vmu": _get_vms_files,
        ".svs": None,
        ".ndpi": None,
        ".scn": None,
        ".bif": None,
    }
    for ext, handler in complex_file_handlers.items():
        complex_files = [file for file in files if file.suffix.lower() == ext]
        if len(complex_files) > 0:
            converted_files, convert_errors = _convert(
                complex_files, handler, converter
            )
            loaded_files += converted_files
            errors.update(convert_errors)

    # don't handle files that are associated files
    for file in files:
        if not any(g.path == file for g in loaded_files) and not any(
            file in g.associated_files
            for g in loaded_files
            if g.associated_files is not None
        ):
            gc_file = GrandChallengeTiffFile(file)
            loaded_files.append(gc_file)
    return loaded_files, errors


def image_builder_tiff(  # noqa: C901
    *, files: Set[Path], output_directory: Path, **_
) -> ImageBuilderResult:
    new_images = set()
    new_image_files = set()
    consumed_files = set()
    invalid_file_errors = {}
    new_folders = set()

    def format_error(message):
        return f"Tiff image builder: {message}"

    loaded_files, errors = _load_gc_files(files=files, converter=pyvips)
    for gc_file in loaded_files:
        dzi_output = None
        error = ""
        if gc_file.path in errors:
            error = errors[gc_file.path]

        # try and load image with tiff file
        try:
            gc_file = _load_with_tiff(gc_file=gc_file)
        except Exception as e:
            error += f"TIFF load error: {e}. "

        # try and load image with open slide
        try:
            gc_file = _load_with_openslide(gc_file=gc_file)
        except Exception as e:
            error += f"OpenSlide load error: {e}. "

        # validate
        try:
            gc_file.validate()
        except ValidationError as e:
            error += f"Validation error: {e}. "
            invalid_file_errors[gc_file.path] = format_error(error)
            continue

        image_out_dir = output_directory / str(gc_file.pk)
        image_out_dir.mkdir()

        image = _create_tiff_image_entry(tiff_file=gc_file)
        new_images.add(image)

        try:
            dzi_output, gc_file = _create_dzi_images(
                gc_file=gc_file, output_directory=image_out_dir
            )
        except ValidationError as e:
            error += f"DZI error: {e}. "

        new_image_files |= _new_image_files(
            gc_file=gc_file, image=image, output_directory=image_out_dir
        )
        new_folders |= _new_folder_uploads(dzi_output=dzi_output, image=image,)

        if gc_file.associated_files:
            consumed_files |= {f for f in gc_file.associated_files}
        else:
            consumed_files.add(gc_file.path)

    return ImageBuilderResult(
        consumed_files=consumed_files,
        file_errors=invalid_file_errors,
        new_images=new_images,
        new_image_files=new_image_files,
        new_folders=new_folders,
    )


def _create_tiff_image_entry(*, tiff_file: GrandChallengeTiffFile) -> PanImg:
    # Builds a new Image model item
    return PanImg(
        pk=tiff_file.pk,
        name=tiff_file.path.name,
        width=tiff_file.image_width,
        height=tiff_file.image_height,
        depth=1,
        resolution_levels=tiff_file.resolution_levels,
        color_space=tiff_file.color_space,
        voxel_width_mm=tiff_file.voxel_width_mm,
        voxel_height_mm=tiff_file.voxel_height_mm,
        voxel_depth_mm=tiff_file.voxel_depth_mm,
        timepoints=None,
        window_center=None,
        window_width=None,
    )


def _create_dzi_images(
    *, gc_file: GrandChallengeTiffFile, output_directory: Path
) -> (Path, GrandChallengeTiffFile):
    # Creates a dzi file(out.dzi) and corresponding tiles in folder {pk}_files
    dzi_output = output_directory / str(gc_file.pk)
    try:
        image = pyvips.Image.new_from_file(
            str(gc_file.path.absolute()), access="sequential"
        )
        pyvips.Image.dzsave(image, str(dzi_output), tile_size=DZI_TILE_SIZE)
        gc_file.source_files.append(
            dzi_output.parent / (dzi_output.name + ".dzi")
        )
    except Exception as e:
        raise ValidationError(f"Image can't be converted to dzi: {e}")

    return dzi_output, gc_file



"""
Image builder for MetaIO mhd/mha files.

See: https://itk.org/Wiki/MetaIO/Documentation
"""

from pathlib import Path
from typing import Mapping, Sequence, Set, Tuple, Union

from panimg.image_builders.metaio_utils import (
    load_sitk_image,
    parse_mh_header,
)
from panimg.image_builders.utils import convert_itk_to_internal
from panimg.models import PanImg, PanImgFile
from panimg.types import ImageBuilderResult


def image_builder_mhd(  # noqa: C901
    *, files: Set[Path], output_directory: Path, **_
) -> ImageBuilderResult:
    """
    Constructs image objects by inspecting files in a directory.

    Parameters
    ----------
    path: Path
        Path to a directory that contains all images that were uploaded during
        an upload session.

    Returns
    -------
    A tuple of
     - all detected images
     - files associated with the detected images
     - path->error message map describing what is wrong with a given file
    """
    element_data_file_key = "ElementDataFile"

    def detect_mhd_file(
        headers: Mapping[str, Union[str, None]], path: Path
    ) -> bool:
        data_file = headers.get(element_data_file_key, None)
        if data_file in [None, "LOCAL"]:
            return False
        data_file_path = (path / Path(data_file)).resolve(strict=False)
        if path not in data_file_path.parents:
            raise ValueError(
                f"{element_data_file_key} references a file which is not in "
                f"the uploaded data folder"
            )
        if not data_file_path.is_file():
            raise ValueError("Data container of mhd file is missing")
        return True

    def detect_mha_file(headers: Mapping[str, Union[str, None]]) -> bool:
        data_file = headers.get(element_data_file_key, None)
        return data_file == "LOCAL"

    def convert_itk_file(
        *, filename: Path, output_dir: Path,
    ) -> Tuple[PanImg, Sequence[PanImgFile]]:
        try:
            simple_itk_image = load_sitk_image(filename.absolute())
        except RuntimeError:
            raise ValueError("SimpleITK cannot open file")

        return convert_itk_to_internal(
            simple_itk_image=simple_itk_image,
            name=filename.name,
            output_directory=output_dir,
        )

    def format_error(message):
        return f"Mhd image builder: {message}"

    new_images = set()
    new_image_files = set()
    consumed_files = set()
    invalid_file_errors = {}
    for file in files:
        try:
            parsed_headers = parse_mh_header(file)
        except ValueError:
            # Maybe add .mhd and .mha files here as "processed" but with errors
            continue

        try:
            is_hd_or_mha = detect_mhd_file(
                parsed_headers, file.parent
            ) or detect_mha_file(parsed_headers)
        except ValueError as e:
            invalid_file_errors[file] = format_error(e)
            continue

        if is_hd_or_mha:
            file_dependency = None
            if parsed_headers[element_data_file_key] != "LOCAL":
                file_dependency = (
                    file.parent / parsed_headers[element_data_file_key]
                )
                if not file_dependency.is_file():
                    invalid_file_errors[file] = format_error(
                        "Cannot find data file"
                    )
                    continue

            n_image, n_image_files = convert_itk_file(
                filename=file, output_dir=output_directory
            )
            new_images.add(n_image)
            new_image_files |= set(n_image_files)

            consumed_files.add(file)
            if file_dependency is not None:
                consumed_files.add(file_dependency)

    return ImageBuilderResult(
        consumed_files=consumed_files,
        file_errors=invalid_file_errors,
        new_images=new_images,
        new_image_files=new_image_files,
        new_folders=set(),
    )



from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Set

from panimg.models import PanImg, PanImgFile, PanImgFolder


@dataclass
class ImageBuilderResult:
    new_images: Set[PanImg]
    new_image_files: Set[PanImgFile]
    new_folders: Set[PanImgFolder]
    consumed_files: Set[Path]
    file_errors: Dict[Path, str]


@dataclass
class PanimgResult(ImageBuilderResult):
    file_errors: Dict[Path, List[str]]



from pathlib import Path
from typing import Set

import SimpleITK
import numpy as np
from PIL import Image
from PIL.Image import DecompressionBombError

from panimg.exceptions import ValidationError
from panimg.image_builders.utils import convert_itk_to_internal
from panimg.types import ImageBuilderResult


def format_error(message):
    return f"Fallback image builder: {message}"


def image_builder_fallback(
    *, files: Set[Path], output_directory: Path, **_
) -> ImageBuilderResult:
    """
    Constructs image objects by inspecting files in a directory.

    Parameters
    ----------
    path
        Path to a directory that contains all images that were uploaded during
        an upload session.

    Returns
    -------
    An `ImageBuilder` object consisting of:
     - a list of filenames for all files consumed by the image builder
     - a list of detected images
     - a list files associated with the detected images
     - path->error message map describing what is wrong with a given file
    """
    errors = {}
    new_images = set()
    new_image_files = set()
    consumed_files = set()
    for file in files:
        try:
            img = Image.open(file)

            if img.format.lower() not in ["jpeg", "png"]:
                raise ValidationError(
                    f"Unsupported image format: {img.format}"
                )

            img_array = np.array(img)
            is_vector = img.mode != "L"
            img = SimpleITK.GetImageFromArray(img_array, isVector=is_vector)
            n_image, n_image_files = convert_itk_to_internal(
                simple_itk_image=img,
                name=file.name,
                use_spacing=False,
                output_directory=output_directory,
            )
            new_images.add(n_image)
            new_image_files |= set(n_image_files)
            consumed_files.add(file)
        except (IOError, ValidationError, DecompressionBombError):
            errors[file] = format_error("Not a valid image file")

    return ImageBuilderResult(
        consumed_files=consumed_files,
        file_errors=errors,
        new_images=new_images,
        new_image_files=new_image_files,
        new_folders=set(),
    )



from collections import defaultdict
from pathlib import Path
from typing import Callable, DefaultDict, Iterable, List, Optional, Set

from panimg.image_builders import DEFAULT_IMAGE_BUILDERS
from panimg.models import PanImg, PanImgFile, PanImgFolder
from panimg.types import ImageBuilderResult, PanimgResult


def convert(
    *,
    input_directory: Path,
    output_directory: Path,
    builders: Optional[Iterable[Callable]] = None,
    created_image_prefix: str = "",
) -> PanimgResult:
    new_images = set()
    new_image_files = set()
    new_folders = set()
    consumed_files = set()
    file_errors = defaultdict(list)

    builders = builders if builders is not None else DEFAULT_IMAGE_BUILDERS

    _convert_directory(
        input_directory=input_directory,
        output_directory=output_directory,
        builders=builders,
        consumed_files=consumed_files,
        new_images=new_images,
        new_image_files=new_image_files,
        new_folders=new_folders,
        file_errors=file_errors,
        created_image_prefix=created_image_prefix,
    )

    return PanimgResult(
        new_images=new_images,
        new_image_files=new_image_files,
        new_folders=new_folders,
        consumed_files=consumed_files,
        file_errors={**file_errors},
    )


def _convert_directory(
    *,
    input_directory: Path,
    output_directory: Path,
    builders: Iterable[Callable],
    consumed_files: Set[Path],
    new_images: Set[PanImg],
    new_image_files: Set[PanImgFile],
    new_folders: Set[PanImgFolder],
    file_errors: DefaultDict[Path, List[str]],
    created_image_prefix: str = "",
):
    input_directory = Path(input_directory).resolve()
    output_directory = Path(output_directory).resolve()

    output_directory.mkdir(exist_ok=True)

    files = set()
    for o in Path(input_directory).iterdir():
        if o.is_dir():
            _convert_directory(
                input_directory=input_directory / o,
                output_directory=output_directory / o,
                builders=builders,
                consumed_files=consumed_files,
                new_images=new_images,
                new_image_files=new_image_files,
                new_folders=new_folders,
                file_errors=file_errors,
                created_image_prefix=created_image_prefix,
            )
        elif o.is_file():
            files.add(o)

    for builder in builders:
        builder_result: ImageBuilderResult = builder(
            files=files - consumed_files,
            output_directory=output_directory,
            created_image_prefix=created_image_prefix,
        )

        new_images |= builder_result.new_images
        new_image_files |= builder_result.new_image_files
        new_folders |= builder_result.new_folders
        consumed_files |= builder_result.consumed_files

        for filepath, msg in builder_result.file_errors.items():
            file_errors[filepath].append(msg)



from panimg.image_builders.dicom import image_builder_dicom
from panimg.image_builders.fallback import image_builder_fallback
from panimg.image_builders.metaio_mhd_mha import image_builder_mhd
from panimg.image_builders.nifti import image_builder_nifti
from panimg.image_builders.tiff import image_builder_tiff

DEFAULT_IMAGE_BUILDERS = [
    image_builder_mhd,
    image_builder_nifti,
    image_builder_dicom,
    image_builder_tiff,
    image_builder_fallback,
]



from pathlib import Path
from typing import Set

import SimpleITK

from panimg.image_builders.utils import convert_itk_to_internal
from panimg.types import ImageBuilderResult


def format_error(message):
    return f"NifTI image builder: {message}"


def image_builder_nifti(
    *, files: Set[Path], output_directory: Path, **_
) -> ImageBuilderResult:
    """
    Constructs image objects from files in NifTI format (nii/nii.gz)

    Parameters
    ----------
    files
        Path to images that were uploaded during an upload session.

    Returns
    -------
    An `ImageBuilder` object consisting of:
     - a list of filenames for all files consumed by the image builder
     - a list of detected images
     - a list files associated with the detected images
     - path->error message map describing what is wrong with a given file
    """
    errors = {}
    new_images = set()
    new_image_files = set()
    consumed_files = set()
    for file in files:
        if not (file.name.endswith(".nii") or file.name.endswith(".nii.gz")):
            continue

        try:
            reader = SimpleITK.ImageFileReader()
            reader.SetImageIO("NiftiImageIO")
            reader.SetFileName(str(file.absolute()))
            img: SimpleITK.Image = reader.Execute()
        except RuntimeError:
            errors[file] = format_error("Not a valid NifTI image file")
            continue

        try:
            n_image, n_image_files = convert_itk_to_internal(
                simple_itk_image=img,
                name=file.name,
                output_directory=output_directory,
            )
            new_images.add(n_image)
            new_image_files |= set(n_image_files)
            consumed_files.add(file)
        except ValueError as e:
            errors[file] = format_error(e)

    return ImageBuilderResult(
        consumed_files=consumed_files,
        file_errors=errors,
        new_images=new_images,
        new_image_files=new_image_files,
        new_folders=set(),
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['panimg', 'panimg.image_builders']

package_data = \
{'': ['*']}

install_requires = \
['Pillow',
 'SimpleITK',
 'numpy',
 'openslide-python',
 'pydantic',
 'pydicom',
 'pyvips',
 'tifffile==2019.1.4']

setup_kwargs = {
    'name': 'panimg',
    'version': '0.1.0',
    'description': 'Conversion of medical images to MHA and TIFF.',
    'long_description': '# panimg\n\n[![CI](https://github.com/DIAGNijmegen/rse-panimg/actions/workflows/ci.yml/badge.svg?branch=main)](https://github.com/DIAGNijmegen/rse-panimg/actions/workflows/ci.yml?query=branch%3Amain)\n![PyPI](https://img.shields.io/pypi/v/panimg)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/panimg)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\n**NOT FOR CLINICAL USE**\n\nConversion of medical images to MHA and TIFF. \nRequires Python 3.7, 3.8 or 3.9.\n`libvips-dev` and `libopenslide-dev` must be installed on your system.\nFor compressed DICOM support ensure that `gdcm` is installed.\n\nUnder the hood we use:\n\n* `SimpleITK`\n* `pydicom`\n* `Pillow`\n* `openslide-python`\n* `pyvips`\n\n## Usage\n\n`panimg` takes a folder full of files and tries to covert them to MHA or TIFF.\nFor each subdirectory of files it will try several strategies for loading the contained files, and if an image is found it will output it to the output folder.\nIt will return a structure containing information about what images were produced, what images were used to form the new images, image metadata, and any errors from any of the strategies.\n\n**NOTE: Alpha software, do not run this on folders you do not have a backup of.**\n\n```python\nfrom pathlib import Path\nfrom panimg import convert\n\nresult = convert(\n    input_directory=Path("/path/to/files/"),\n    output_directory=Path("/where/files/will/go/"),\n)\n```\n\n### Supported Formats\n\n| Input                               | Output           | Strategy   | Notes                      |\n| ----------------------------------- | ---------------- | ---------- | -------------------------- |\n| `.mha`                              | `.mha`           | `metaio`   |                            |\n| `.mhd` with `.raw` or `.zraw`       | `.mha`           | `metaio`   |                            |\n| `.dcm`                              | `.mha`           | `dicom`    | <sup>[1](#footnote1)</sup> |\n| `.nii`                              | `.mha`           | `nifty`    |                            |\n| `.nii.gz`                           | `.mha`           | `nifty`    |                            |\n| `.png`                              | `.mha`           | `fallback` | <sup>[2](#footnote2)</sup> |\n| `.jpeg`                             | `.mha`           | `fallback` | <sup>[2](#footnote2)</sup> |\n| `.tiff`                             | `.tiff` & `.dzi` | `tiff`     | <sup>[3](#footnote3)</sup> |\n| `.svs` (Aperio)                     | `.tiff` & `.dzi` | `tiff`     | <sup>[3](#footnote3)</sup> |\n| `.vms`, `.vmu`, `.ndpi` (Hamamatsu) | `.tiff` & `.dzi` | `tiff`     | <sup>[3](#footnote3)</sup> |\n| `.scn` (Leica)                      | `.tiff` & `.dzi` | `tiff`     | <sup>[3](#footnote3)</sup> |\n| `.mrxs` (MIRAX)                     | `.tiff` & `.dzi` | `tiff`     | <sup>[3](#footnote3)</sup> |\n| `.biff` (Ventana)                   | `.tiff` & `.dzi` | `tiff`     | <sup>[3](#footnote3)</sup> |\n\n<a name="footnote1">1</a>: Compressed DICOM requires `gdcm`\n\n<a name="footnote2">2</a>: 2D only, unitary dimensions\n\n<a name="footnote3">3</a>: DZI only created if possible\n',
    'author': 'James Meakin',
    'author_email': '12661555+jmsmkn@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/DIAGNijmegen/rse-panimg',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from pathlib import Path
from typing import AnyStr, Optional, Sequence, Tuple
from uuid import uuid4

import SimpleITK

from panimg.models import ColorSpace, ImageType, PanImg, PanImgFile
from panimg.settings import ITK_INTERNAL_FILE_FORMAT


def convert_itk_to_internal(
    *,
    simple_itk_image: SimpleITK.Image,
    output_directory: Path,
    name: Optional[AnyStr] = None,
    use_spacing: Optional[bool] = True,
) -> Tuple[PanImg, Sequence[PanImgFile]]:
    color_space = simple_itk_image.GetNumberOfComponentsPerPixel()
    color_space = {
        1: ColorSpace.GRAY,
        3: ColorSpace.RGB,
        4: ColorSpace.RGBA,
    }.get(color_space, None)
    if color_space is None:
        raise ValueError("Unknown color space for MetaIO image.")

    pk = uuid4()
    if not name:
        name = str(pk)

    work_dir = Path(output_directory) / str(pk)
    work_dir.mkdir()

    SimpleITK.WriteImage(
        simple_itk_image,
        str(work_dir / f"{pk}.{ITK_INTERNAL_FILE_FORMAT}"),
        True,
    )

    if simple_itk_image.GetDimension() == 4:
        timepoints = simple_itk_image.GetSize()[-1]
    else:
        timepoints = None
    depth = simple_itk_image.GetDepth()

    try:
        window_center = float(simple_itk_image.GetMetaData("WindowCenter"))
    except (RuntimeError, ValueError):
        window_center = None
    try:
        window_width = float(simple_itk_image.GetMetaData("WindowWidth"))
    except (RuntimeError, ValueError):
        window_width = None

    db_image = PanImg(
        pk=pk,
        name=name,
        width=simple_itk_image.GetWidth(),
        height=simple_itk_image.GetHeight(),
        depth=depth if depth else None,
        window_center=window_center,
        window_width=window_width,
        timepoints=timepoints,
        resolution_levels=None,
        color_space=color_space,
        voxel_width_mm=simple_itk_image.GetSpacing()[0]
        if use_spacing
        else None,
        voxel_height_mm=simple_itk_image.GetSpacing()[1]
        if use_spacing
        else None,
        voxel_depth_mm=simple_itk_image.GetSpacing()[2] if depth else None,
    )

    db_image_files = []
    for file in work_dir.iterdir():
        db_image_file = PanImgFile(
            image_id=db_image.pk, image_type=ImageType.MHD, file=file,
        )
        db_image_files.append(db_image_file)

    return db_image, db_image_files



from enum import Enum
from pathlib import Path
from typing import Optional
from uuid import UUID

from pydantic.dataclasses import dataclass


class ColorSpace(str, Enum):
    GRAY = "GRAY"
    RGB = "RGB"
    RGBA = "RGBA"
    YCBCR = "YCBCR"


class ImageType(str, Enum):
    MHD = "MHD"
    TIFF = "TIFF"
    DZI = "DZI"


@dataclass(frozen=True)
class PanImg:
    pk: UUID
    name: str
    width: int
    height: int
    depth: Optional[int]
    voxel_width_mm: Optional[float]
    voxel_height_mm: Optional[float]
    voxel_depth_mm: Optional[float]
    timepoints: Optional[int]
    resolution_levels: Optional[int]
    window_center: Optional[float]
    window_width: Optional[float]
    color_space: ColorSpace


@dataclass(frozen=True)
class PanImgFile:
    image_id: UUID
    image_type: ImageType
    file: Path


@dataclass(frozen=True)
class PanImgFolder:
    image_id: UUID
    folder: Path
